@@ -22,6 +22,7 @@ from .. import mesonlib
 from ..environment import detect_cpu_family
 
 from .base import (DependencyException, ExternalDependency)
+from .misc import ThreadDependency
 
 # On windows 3 directory layouts are supported:
 # * The default layout (versioned) installed:
@@ -63,6 +64,7 @@ from .base import (DependencyException, ExternalDependency)
 # Win   / Cygwin:   libboost_<module>.dll.a                                 (location = /usr/lib)
 #                   libboost_<module>.a
 #                   cygboost_<module>_1_64.dll                              (location = /usr/bin)
+# Win   / VS:       boost_<module>-vc<ver>-mt[-gd]-<arch>-1_67.dll          (location = C:/local/boost_1_67_0)
 # Mac   / homebrew: libboost_<module>.dylib + libboost_<module>-mt.dylib    (location = /usr/local/lib)
 # Mac   / macports: libboost_<module>.dylib + libboost_<module>-mt.dylib    (location = /opt/local/lib)
 #
@@ -102,6 +104,8 @@ class BoostDependency(ExternalDependency):
         self.is_multithreading = threading == "multi"
 
         self.requested_modules = self.get_requested(kwargs)
+        if 'thread' in self.requested_modules:
+            self._add_sub_dependency(ThreadDependency, environment, kwargs)
 
         self.boost_root = None
         self.boost_roots = []
@@ -119,19 +123,18 @@ class BoostDependency(ExternalDependency):
             self.libdir = os.environ['BOOST_LIBRARYDIR']
 
         if self.boost_root is None:
-            if mesonlib.for_windows(self.want_cross, self.env):
+            if self.env.machines[self.for_machine].is_windows():
                 self.boost_roots = self.detect_win_roots()
             else:
                 self.boost_roots = self.detect_nix_roots()
 
         if self.incdir is None:
-            if mesonlib.for_windows(self.want_cross, self.env):
+            if self.env.machines[self.for_machine].is_windows():
                 self.incdir = self.detect_win_incdir()
             else:
                 self.incdir = self.detect_nix_incdir()
 
         if self.check_invalid_modules():
-            self.log_fail()
             return
 
         mlog.debug('Boost library root dir is', mlog.bold(self.boost_root))
@@ -145,12 +148,6 @@ class BoostDependency(ExternalDependency):
             self.detect_lib_modules()
             mlog.debug('Boost library directory is', mlog.bold(self.libdir))
 
-        # 3. Report success or failure
-        if self.is_found:
-            self.log_success()
-        else:
-            self.log_fail()
-
     def check_invalid_modules(self):
         invalid_modules = [c for c in self.requested_modules if 'boost_' + c not in BOOST_LIBS]
 
@@ -171,17 +168,14 @@ class BoostDependency(ExternalDependency):
         else:
             return False
 
-    def log_fail(self):
+    def log_details(self):
         module_str = ', '.join(self.requested_modules)
-        mlog.log("Dependency Boost (%s) found:" % module_str, mlog.red('NO'))
+        return module_str
 
-    def log_success(self):
-        module_str = ', '.join(self.requested_modules)
+    def log_info(self):
         if self.boost_root:
-            info = self.version + ', ' + self.boost_root
-        else:
-            info = self.version
-        mlog.log('Dependency Boost (%s) found:' % module_str, mlog.green('YES'), info)
+            return self.boost_root
+        return ''
 
     def detect_nix_roots(self):
         return [os.path.abspath(os.path.join(x, '..'))
@@ -215,7 +209,7 @@ class BoostDependency(ExternalDependency):
         for root in self.boost_roots:
             globtext = os.path.join(root, 'include', 'boost-*')
             incdirs = glob.glob(globtext)
-            if len(incdirs) > 0:
+            if incdirs:
                 return incdirs[0]
             incboostdir = os.path.join(root, 'include', 'boost')
             if os.path.isdir(incboostdir):
@@ -256,7 +250,7 @@ class BoostDependency(ExternalDependency):
 
     def detect_headers_and_version(self):
         try:
-            version = self.clib_compiler.get_define('BOOST_LIB_VERSION', '#include <boost/version.hpp>', self.env, self.get_compile_args(), [])
+            version = self.clib_compiler.get_define('BOOST_LIB_VERSION', '#include <boost/version.hpp>', self.env, self.get_compile_args(), [], disable_cache=True)[0]
         except mesonlib.EnvironmentException:
             return
         except TypeError:
@@ -269,13 +263,12 @@ class BoostDependency(ExternalDependency):
 
     def detect_lib_modules(self):
         self.lib_modules = {}
-
         # 1. Try to find modules using compiler.find_library( )
         if self.find_libraries_with_abi_tags(self.abi_tags()):
             pass
         # 2. Fall back to the old method
         else:
-            if mesonlib.for_windows(self.want_cross, self.env):
+            if self.env.machines[self.for_machine].is_windows():
                 self.detect_lib_modules_win()
             else:
                 self.detect_lib_modules_nix()
@@ -296,9 +289,9 @@ class BoostDependency(ExternalDependency):
 
     def compiler_tag(self):
         tag = None
-        compiler = self.env.detect_cpp_compiler(self.want_cross)
-        if mesonlib.for_windows(self.want_cross, self.env):
-            if compiler.get_id() == 'msvc':
+        compiler = self.env.detect_cpp_compiler(self.for_machine)
+        if self.env.machines[self.for_machine].is_windows():
+            if compiler.get_id() in ['msvc', 'clang-cl']:
                 comp_ts_version = compiler.get_toolset_version()
                 compiler_ts = comp_ts_version.split('.')
                 # FIXME - what about other compilers?
@@ -311,10 +304,10 @@ class BoostDependency(ExternalDependency):
         if not self.is_multithreading:
             return ''
 
-        if mesonlib.for_darwin(self.want_cross, self.env):
+        if self.env.machines[self.for_machine].is_darwin():
             # - Mac:      requires -mt for multithreading, so should not fall back to non-mt libraries.
             return '-mt'
-        elif mesonlib.for_windows(self.want_cross, self.env):
+        elif self.env.machines[self.for_machine].is_windows():
             # - Windows:  requires -mt for multithreading, so should not fall back to non-mt libraries.
             return '-mt'
         else:
@@ -328,21 +321,35 @@ class BoostDependency(ExternalDependency):
     def debug_tag(self):
         return '-gd' if self.is_debug else ''
 
+    def arch_tag(self):
+        # currently only applies to windows msvc installed binaries
+        if self.env.detect_cpp_compiler(self.for_machine).get_id() not in ['msvc', 'clang-cl']:
+            return ''
+        # pre-compiled binaries only added arch tag for versions > 1.64
+        if float(self.version) < 1.65:
+            return ''
+        arch = detect_cpu_family(self.env.coredata.compilers.host)
+        if arch == 'x86':
+            return '-x32'
+        elif arch == 'x86_64':
+            return '-x64'
+        return ''
+
     def versioned_abi_tag(self):
-        return self.compiler_tag() + self.threading_tag() + self.debug_tag() + self.version_tag()
+        return self.compiler_tag() + self.threading_tag() + self.debug_tag() + self.arch_tag() + self.version_tag()
 
     # FIXME - how to handle different distributions, e.g. for Mac? Currently we handle homebrew and macports, but not fink.
     def abi_tags(self):
-        if mesonlib.for_windows(self.want_cross, self.env):
+        if self.env.machines[self.for_machine].is_windows():
             return [self.versioned_abi_tag(), self.threading_tag()]
         else:
             return [self.threading_tag()]
 
     def sourceforge_dir(self):
-        if self.env.detect_cpp_compiler(self.want_cross).get_id() != 'msvc':
+        if self.env.detect_cpp_compiler(self.for_machine).get_id() != 'msvc':
             return None
-        comp_ts_version = self.env.detect_cpp_compiler(self.want_cross).get_toolset_version()
-        arch = detect_cpu_family(self.env.coredata.compilers)
+        comp_ts_version = self.env.detect_cpp_compiler(self.for_machine).get_toolset_version()
+        arch = detect_cpu_family(self.env.coredata.compilers.host)
         if arch == 'x86':
             return 'lib32-msvc-{}'.format(comp_ts_version)
         elif arch == 'x86_64':
@@ -420,7 +427,7 @@ class BoostDependency(ExternalDependency):
         for entry in globber2_matches:
             fname = os.path.basename(entry)
             self.lib_modules[self.modname_from_filename(fname)] = [fname]
-        if len(globber2_matches) == 0:
+        if not globber2_matches:
             # FIXME - why are we only looking for *.lib? Mingw provides *.dll.a and *.a
             for entry in glob.glob(os.path.join(self.libdir, globber1 + '.lib')):
                 if self.static:
@@ -430,7 +437,7 @@ class BoostDependency(ExternalDependency):
     def detect_lib_modules_nix(self):
         if self.static:
             libsuffix = 'a'
-        elif mesonlib.for_darwin(self.want_cross, self.env):
+        elif self.env.machines[self.for_machine].is_darwin():
             libsuffix = 'dylib'
         else:
             libsuffix = 'so'
@@ -484,9 +491,6 @@ class BoostDependency(ExternalDependency):
     def get_sources(self):
         return []
 
-    def need_threads(self):
-        return 'thread' in self.requested_modules
-
 
 # Generated with boost_names.py
 BOOST_LIBS = [
