@@ -65,7 +65,7 @@ class DirMaker:
     def __enter__(self):
         return self
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, exception_type, value, traceback):
         self.dirs.reverse()
         for d in self.dirs:
             append_to_log(self.lf, d)
@@ -93,7 +93,7 @@ def set_chown(path, user=None, group=None, dir_fd=None, follow_symlinks=True):
                                                  dir_fd=dir_fd,
                                                  follow_symlinks=follow_symlinks)
         shutil.chown(path, user, group)
-    except:
+    except Exception:
         raise
     finally:
         os.chown = real_os_chown
